UNKNOWN = "unknown"

# algorand
ALGOXNFT_COM = "algoxnft.com"
RANDGALLERY_COM = "randgallery.com"
SHUFL_APP = "shufl.app"

# cardano
JPG_STORE = "jpg.store"

# evm
OPENSEA_IO = "opensea.io"


MARKETPLACES = [
    ALGOXNFT_COM,
    RANDGALLERY_COM,
    SHUFL_APP,
    JPG_STORE,
    OPENSEA_IO,
]
